@@ -18,6 +18,8 @@
    #define speller manager
    #define aspell_speller_check pspell_manager_check
    #define aspell_speller_add_to_session pspell_manager_add_to_session
+   #define aspell_speller_add_to_personal pspell_manager_add_to_personal
+   #define aspell_speller_save_all_word_lists pspell_manager_save_all_word_lists
    #define aspell_speller_store_replacement pspell_manager_store_replacement
    #define AspellWordList PspellWordList
    #define AspellStringEnumeration PspellStringEmulation
@@ -44,8 +46,11 @@ const int quiet = 0;
 struct _GtkSpell {
 	GtkTextView *view;
 	GtkTextTag *tag_highlight;
-	GtkTextMark *mark_insert;
+	GtkTextMark *mark_insert_start;
+	GtkTextMark *mark_insert_end;
+	gboolean deferred_check;
 	AspellSpeller *speller;
+	GtkTextMark *mark_click;
 };
 
 static void gtkspell_free(GtkSpell *spell);
@@ -115,8 +120,9 @@ check_word(GtkSpell *spell, GtkTextBuffer *buffer,
 	char *text;
 	text = gtk_text_buffer_get_text(buffer, start, end, FALSE);
 	if (debug) g_print("checking: %s\n", text);
-	if (aspell_speller_check(spell->speller, text, -1) == FALSE)
-		gtk_text_buffer_apply_tag(buffer, spell->tag_highlight, start, end);
+	if (g_unichar_isdigit(*text) == FALSE) /* don't check numbers */
+		if (aspell_speller_check(spell->speller, text, -1) == FALSE)
+			gtk_text_buffer_apply_tag(buffer, spell->tag_highlight, start, end);
 	g_free(text);
 }
 
@@ -130,12 +136,13 @@ print_iter(char *name, GtkTextIter *iter) {
 
 static void
 check_range(GtkSpell *spell, GtkTextBuffer *buffer,
-            GtkTextIter start, GtkTextIter end) {
+            GtkTextIter start, GtkTextIter end, gboolean force_all) {
 	/* we need to "split" on word boundaries.
 	 * luckily, pango knows what "words" are 
 	 * so we don't have to figure it out. */
 
-	GtkTextIter wstart, wend;
+	GtkTextIter wstart, wend, cursor, precursor;
+	gboolean inword, highlight;
 	if (debug) {
 		g_print("check_range: "); print_iter("s", &start); print_iter("e", &end); g_print(" -> ");
 	}
@@ -155,8 +162,25 @@ check_range(GtkSpell *spell, GtkTextBuffer *buffer,
 				gtk_text_iter_backward_word_start(&start);
 		}
 	}
+	gtk_text_buffer_get_iter_at_mark(buffer, &cursor,
+			gtk_text_buffer_get_insert(buffer));
+
+	precursor = cursor;
+	gtk_text_iter_backward_char(&precursor);
+	highlight = gtk_text_iter_has_tag(&cursor, spell->tag_highlight) ||
+			gtk_text_iter_has_tag(&precursor, spell->tag_highlight);
+	
 	gtk_text_buffer_remove_tag(buffer, spell->tag_highlight, &start, &end);
 
+	/* Fix a corner case when replacement occurs at beginning of buffer:
+	 * An iter at offset 0 seems to always be inside a word,
+	 * even if it's not.  Possibly a pango bug.
+	 */
+	if (gtk_text_iter_get_offset(&start) == 0) {
+		gtk_text_iter_forward_word_end(&start);
+		gtk_text_iter_backward_word_start(&start);
+	}
+
 	if (debug) {print_iter("s", &start); print_iter("e", &end); g_print("\n");}
 
 	wstart = start;
@@ -165,7 +189,21 @@ check_range(GtkSpell *spell, GtkTextBuffer *buffer,
 		wend = wstart;
 		gtk_text_iter_forward_word_end(&wend);
 
-		check_word(spell, buffer, &wstart, &wend);
+		inword = (gtk_text_iter_compare(&wstart, &cursor) < 0) && 
+				(gtk_text_iter_compare(&cursor, &wend) <= 0);
+
+		if (inword && !force_all) {
+			/* this word is being actively edited, 
+			 * only check if it's already highligted,
+			 * otherwise defer this check until later. */
+			if (highlight)
+				check_word(spell, buffer, &wstart, &wend);
+			else
+				spell->deferred_check = TRUE;
+		} else {
+			check_word(spell, buffer, &wstart, &wend);
+			spell->deferred_check = FALSE;
+		}
 
 		/* now move wend to the beginning of the next word, */
 		gtk_text_iter_forward_word_end(&wend);
@@ -179,6 +217,14 @@ check_range(GtkSpell *spell, GtkTextBuffer *buffer,
 	}
 }
 
+static void
+check_deferred_range(GtkSpell *spell, GtkTextBuffer *buffer, gboolean force_all) {
+	GtkTextIter start, end;
+	gtk_text_buffer_get_iter_at_mark(buffer, &start, spell->mark_insert_start);
+	gtk_text_buffer_get_iter_at_mark(buffer, &end, spell->mark_insert_end);
+	check_range(spell, buffer, start, end, force_all);
+}
+
 /* insertion works like this:
  *  - before the text is inserted, we mark the position in the buffer.
  *  - after the text is inserted, we see where our mark is and use that and
@@ -189,7 +235,7 @@ check_range(GtkSpell *spell, GtkTextBuffer *buffer,
 static void
 insert_text_before(GtkTextBuffer *buffer, GtkTextIter *iter,
                    gchar *text, gint len, GtkSpell *spell) {
-	gtk_text_buffer_move_mark(buffer, spell->mark_insert, iter);
+	gtk_text_buffer_move_mark(buffer, spell->mark_insert_start, iter);
 }
 
 static void
@@ -197,9 +243,13 @@ insert_text_after(GtkTextBuffer *buffer, GtkTextIter *iter,
                   gchar *text, gint len, GtkSpell *spell) {
 	GtkTextIter start;
 
+	if (debug) g_print("insert\n");
+
 	/* we need to check a range of text. */
-	gtk_text_buffer_get_iter_at_mark(buffer, &start, spell->mark_insert);
-	check_range(spell, buffer, start, *iter);
+	gtk_text_buffer_get_iter_at_mark(buffer, &start, spell->mark_insert_start);
+	check_range(spell, buffer, start, *iter, FALSE);
+	
+	gtk_text_buffer_move_mark(buffer, spell->mark_insert_end, iter);
 }
 
 /* deleting is more simple:  we're given the range of deleted text.
@@ -212,14 +262,22 @@ insert_text_after(GtkTextBuffer *buffer, GtkTextIter *iter,
 static void
 delete_range_after(GtkTextBuffer *buffer,
                    GtkTextIter *start, GtkTextIter *end, GtkSpell *spell) {
-	check_range(spell, buffer, *start, *end);
+	if (debug) g_print("delete\n");
+	check_range(spell, buffer, *start, *end, FALSE);
 }
 
 static void
-get_cur_word_extents(GtkTextBuffer *buffer,
-                     GtkTextIter *start, GtkTextIter *end) {
-	gtk_text_buffer_get_iter_at_mark(buffer, start, 
-			gtk_text_buffer_get_insert(buffer));
+mark_set(GtkTextBuffer *buffer, GtkTextIter *iter, 
+		 GtkTextMark *mark, GtkSpell *spell) {
+	/* if the cursor has moved and there is a deferred check so handle it now */
+	if ((mark == gtk_text_buffer_get_insert(buffer)) && spell->deferred_check)
+		check_deferred_range(spell, buffer, FALSE);
+}
+
+static void
+get_word_extents_from_mark(GtkTextBuffer *buffer,
+                     GtkTextIter *start, GtkTextIter *end, GtkTextMark *mark) {
+	gtk_text_buffer_get_iter_at_mark(buffer, start, mark);
 	if (!gtk_text_iter_starts_word(start)) 
 		gtk_text_iter_backward_word_start(start);
 	*end = *start;
@@ -235,12 +293,31 @@ add_to_dictionary(GtkWidget *menuitem, GtkSpell *spell) {
 	
 	buffer = gtk_text_view_get_buffer(spell->view);
 
-	get_cur_word_extents(buffer, &start, &end);
+	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
+	word = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
+	
+	aspell_speller_add_to_personal(spell->speller, word, strlen(word));
+	aspell_speller_save_all_word_lists(spell->speller);
+
+	gtkspell_recheck_all(spell);
+
+	g_free(word);
+}
+
+static void
+ignore_all(GtkWidget *menuitem, GtkSpell *spell) {
+	char *word;
+	GtkTextIter start, end;
+	GtkTextBuffer *buffer;
+	
+	buffer = gtk_text_view_get_buffer(spell->view);
+
+	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
 	word = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
 	
 	aspell_speller_add_to_session(spell->speller, word, strlen(word));
 
-	gtk_text_buffer_remove_tag(buffer, spell->tag_highlight, &start, &end);
+	gtkspell_recheck_all(spell);
 
 	g_free(word);
 }
@@ -254,7 +331,7 @@ replace_word(GtkWidget *menuitem, GtkSpell *spell) {
 	
 	buffer = gtk_text_view_get_buffer(spell->view);
 
-	get_cur_word_extents(buffer, &start, &end);
+	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
 	oldword = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
 	newword = gtk_label_get_text(GTK_LABEL(GTK_BIN(menuitem)->child));
 
@@ -289,22 +366,6 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 	
 	topmenu = menu = gtk_menu_new();
 
-	/* + Add to Dictionary */
-	label = g_strdup_printf("Add \"%s\" to Dictionary", word);
-	mi = gtk_image_menu_item_new_with_label(label);
-	g_free(label);
-	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), 
-			gtk_image_new_from_stock(GTK_STOCK_ADD, GTK_ICON_SIZE_MENU));
-	g_signal_connect(G_OBJECT(mi), "activate",
-			G_CALLBACK(add_to_dictionary), spell);
-	gtk_widget_show_all(mi);
-	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
-
-	/* Separator */
-	mi = gtk_menu_item_new();
-	gtk_widget_show(mi);
-	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
-
 	suggestions = aspell_speller_suggest(spell->speller, word, -1);
 	elements = aspell_word_list_elements(suggestions);
 
@@ -347,6 +408,31 @@ build_suggestion_menu(GtkSpell *spell, GtkTextBuffer *buffer,
 
 	delete_aspell_string_enumeration(elements);
 
+	/* Separator */
+	mi = gtk_menu_item_new();
+	gtk_widget_show(mi);
+	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
+
+	/* + Add to Dictionary */
+	label = g_strdup_printf("Add \"%s\" to Dictionary", word);
+	mi = gtk_image_menu_item_new_with_label(label);
+	g_free(label);
+	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), 
+			gtk_image_new_from_stock(GTK_STOCK_ADD, GTK_ICON_SIZE_MENU));
+	g_signal_connect(G_OBJECT(mi), "activate",
+			G_CALLBACK(add_to_dictionary), spell);
+	gtk_widget_show_all(mi);
+	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
+
+	/* - Ignore All */
+	mi = gtk_image_menu_item_new_with_label("Ignore All");
+	gtk_image_menu_item_set_image(GTK_IMAGE_MENU_ITEM(mi), 
+			gtk_image_new_from_stock(GTK_STOCK_REMOVE, GTK_ICON_SIZE_MENU));
+	g_signal_connect(G_OBJECT(mi), "activate",
+			G_CALLBACK(ignore_all), spell);
+	gtk_widget_show_all(mi);
+	gtk_menu_shell_append(GTK_MENU_SHELL(topmenu), mi);
+
 	return topmenu;
 }
 
@@ -358,7 +444,7 @@ populate_popup(GtkTextView *textview, GtkMenu *menu, GtkSpell *spell) {
 	char *word;
 
 	/* we need to figure out if they picked a misspelled word. */
-	get_cur_word_extents(buffer, &start, &end);
+	get_word_extents_from_mark(buffer, &start, &end, spell->mark_click);
 
 	/* if our highlight algorithm ever messes up, 
 	 * this isn't correct, either. */
@@ -385,22 +471,25 @@ populate_popup(GtkTextView *textview, GtkMenu *menu, GtkSpell *spell) {
 }
 
 /* when the user right-clicks on a word, they want to check that word.
- * here, we move the cursor to the location of the clicked-upon word.
- * is this necessary?  we could maybe just check the word from its
- * existing location... */
+ * here, we do NOT  move the cursor to the location of the clicked-upon word
+ * since that prevents the use of edit functions on the context menu. */
 static gboolean
-button_press_event(GtkTextView *view, GdkEventButton *event, gpointer data) {
+button_press_event(GtkTextView *view, GdkEventButton *event, GtkSpell *spell) {
 	if (event->button == 3) {
 		gint x, y;
 		GtkTextIter iter;
+		GtkTextBuffer *buffer = gtk_text_view_get_buffer(view);
+
+		/* handle deferred check if it exists */
+		if (spell->deferred_check)
+			check_deferred_range(spell, buffer, TRUE);
 
 		gtk_text_view_window_to_buffer_coords(view, 
 				GTK_TEXT_WINDOW_TEXT, 
 				event->x, event->y,
 				&x, &y);
 		gtk_text_view_get_iter_at_location(view, &iter, x, y);
-		gtk_text_buffer_place_cursor(gtk_text_view_get_buffer(view),
-				&iter);
+		gtk_text_buffer_move_mark(buffer, spell->mark_click, &iter);
 	}
 	return FALSE; /* false: let gtk process this event, too.
 					 we don't want to eat any events. */
@@ -487,7 +576,7 @@ gtkspell_recheck_all(GtkSpell *spell) {
 
 	gtk_text_buffer_get_bounds(buffer, &start, &end);
 
-	check_range(spell, buffer, start, end);
+	check_range(spell, buffer, start, end, TRUE);
 }
 
 /**
@@ -541,6 +630,9 @@ gtkspell_new_attach(GtkTextView *view, const gchar *lang, GError **error) {
 	g_signal_connect_after(G_OBJECT(buffer),
 			"delete-range",
 			G_CALLBACK(delete_range_after), spell);
+	g_signal_connect(G_OBJECT(buffer),
+			"mark-set",
+			G_CALLBACK(mark_set), spell);
 
 	spell->tag_highlight = gtk_text_buffer_create_tag(buffer,
 			"gtkspell-misspelled",
@@ -551,9 +643,17 @@ gtkspell_new_attach(GtkTextView *view, const gchar *lang, GError **error) {
 	/* we create the mark here, but we don't use it until text is
 	 * inserted, so we don't really care where iter points.  */
 	gtk_text_buffer_get_bounds(buffer, &start, &end);
-	spell->mark_insert = gtk_text_buffer_create_mark(buffer,
-			"gtkspell-insert",
+	spell->mark_insert_start = gtk_text_buffer_create_mark(buffer,
+			"gtkspell-insert-start",
 			&start, TRUE);
+	spell->mark_insert_end = gtk_text_buffer_create_mark(buffer,
+			"gtkspell-insert-end",
+			&start, TRUE);
+	spell->mark_click = gtk_text_buffer_create_mark(buffer,
+			"gtkspell-click",
+			&start, TRUE);
+		
+	spell->deferred_check = FALSE;
 
 	/* now check the entire text buffer. */
 	gtkspell_recheck_all(spell);
@@ -573,7 +673,9 @@ gtkspell_free(GtkSpell *spell) {
 	gtk_text_buffer_remove_tag(buffer, spell->tag_highlight, &start, &end);
 	gtk_text_tag_table_remove(table, spell->tag_highlight);
 
-	gtk_text_buffer_delete_mark(buffer, spell->mark_insert);
+	gtk_text_buffer_delete_mark(buffer, spell->mark_insert_start);
+	gtk_text_buffer_delete_mark(buffer, spell->mark_insert_end);
+	gtk_text_buffer_delete_mark(buffer, spell->mark_click);
 
 	delete_aspell_speller(spell->speller);
 
